@@ -90,9 +90,13 @@ class _ReadLine(object):
             pass
 
     class History(object):
-        def __init__(self):
+        def __init__(self, initial_history):
             object.__init__(self)
-            self.items = ['']
+            if initial_history is None or len(initial_history) == 0:
+              initial_history = ['']
+            elif initial_history[0] != '':
+              initial_history = [''] + initial_history
+            self.items = initial_history
             self.ptr = 0
             self.edited = {}
 
@@ -111,10 +115,16 @@ class _ReadLine(object):
             elif self.ptr in self.edited:
                 del self.edited[self.ptr]
 
-            self.ptr = self.ptr + dir
-            if self.ptr >= len(self.items):
-                self.ptr = 0
-            elif self.ptr < 0:
+            # Do not cycle. It's confusing.
+            if (self.ptr == 0 and dir > 0) or self.ptr + dir >= len(self.items):
+              # Position 0 is a bit weird. It's the last position, not
+              # the first!
+              self.ptr = 0
+            elif self.ptr > 0 and dir < 0 and self.ptr + dir < 1:
+              self.ptr = 1
+            else:
+              self.ptr = self.ptr + dir
+              if self.ptr < 0:
                 self.ptr = len(self.items) - 1
 
             try:
@@ -122,7 +132,47 @@ class _ReadLine(object):
             except KeyError:
                 return self.items[self.ptr]
 
-    def __init__(self, quit_func=None):
+        def search_backward(self, text):
+            if len(self.items) == 1:
+              return None, None
+
+            if self.ptr == 1:
+              return None, None
+
+            if self.ptr == 0:
+              search_iter = len(self.items) - 1
+            else:
+              search_iter = self.ptr - 1
+
+            while text not in self.items[search_iter]:
+              search_iter -= 1
+              if search_iter == 1:
+                return None, self.items[1]
+
+            return search_iter, self.items[search_iter]
+
+        def search_forward(self, text):
+            if len(self.items) == 1:
+              return None, None
+
+            if self.ptr == 0:
+              return None, None
+
+            if self.ptr == len(self.items) - 1:
+              search_iter = 0
+            else:
+              search_iter = self.ptr + 1
+
+            while text not in self.items[search_iter]:
+              if search_iter == 0:
+                return None, None
+              search_iter += 1
+              if search_iter > len(self.items):
+                search_iter = 0
+
+            return search_iter, self.items[search_iter]
+
+    def __init__(self, quit_func=None, initial_history=None):
         object.__init__(self)
 
         self.quit_func = quit_func
@@ -137,8 +187,22 @@ class _ReadLine(object):
         self.do_insert = False
         self.do_delete = False
 
-        self.stdout_tag = self.buffer.create_tag("stdout", foreground="#006000")
-        self.stderr_tag = self.buffer.create_tag("stderr", foreground="#B00000")
+        #Use the theme's color scheme for the text color
+        font_color = self.get_style_context().get_property('color', Gtk.StateFlags.NORMAL)
+        r = int (font_color.red * 255)
+        g = int (font_color.green * 255)
+        b = int (font_color.blue * 255)
+        hex_font_color = "#" + '{r:02x}{g:02x}{b:02x}'.format (r = r, g = g, b = b)
+
+        #Choose an error text color based on background luminance
+        bg_color = self.get_style_context().get_property('background-color', Gtk.StateFlags.NORMAL)
+        lum = (bg_color.red * 0.22248840) + (bg_color.green * 0.71690369) + (bg_color.blue * 0.06060791)
+        background_color = "#B00000"
+        if (lum < 0.4):
+          background_color = "#FF6161"
+
+        self.stdout_tag = self.buffer.create_tag("stdout", foreground=hex_font_color)
+        self.stderr_tag = self.buffer.create_tag("stderr", foreground=background_color)
         self._stdout = _ReadLine.Output(self, "stdout")
         self._stderr = _ReadLine.Output(self, "stderr")
 
@@ -154,8 +218,11 @@ class _ReadLine(object):
         self.in_modal_raw_input = False
         self.run_on_raw_input = None
         self.tab_pressed = 0
-        self.history = _ReadLine.History()
-        self.nonword_re = re.compile("[^\w\._]")
+        self.history = _ReadLine.History(initial_history)
+        self.nonword_re = re.compile(r"[^\w\._]")
+
+        self._searching_backward = False
+        self._searching_forward  = False
 
     def freeze_undo(self):
         try: self.begin_not_undoable_action()
@@ -291,29 +358,36 @@ class _ReadLine(object):
         self.tab_pressed = 0
         handled = True
 
-        state = event.state & (Gdk.ModifierType.SHIFT_MASK |
-                               Gdk.ModifierType.CONTROL_MASK |
+        state = event.state & (Gdk.ModifierType.CONTROL_MASK |
                                Gdk.ModifierType.MOD1_MASK)
         keyval = event.keyval
 
         if not state:
             if keyval == Gdk.KEY_Escape:
-                pass
+                self.__history_search_stop()
             elif keyval == Gdk.KEY_Return:
+                self.__history_search_stop()
                 self._commit()
-            elif keyval == Gdk.KEY_Up:
+            elif keyval == Gdk.KEY_Up or keyval == Gdk.KEY_Page_Up:
+                self.__history_search_stop()
                 self.__history(-1)
-            elif keyval == Gdk.KEY_Down:
+            elif keyval == Gdk.KEY_Down or keyval == Gdk.KEY_Page_Down:
+                self.__history_search_stop()
                 self.__history(1)
             elif keyval == Gdk.KEY_Left:
+                self.__history_search_stop()
                 self.__move_cursor(-1)
             elif keyval == Gdk.KEY_Right:
+                self.__history_search_stop()
                 self.__move_cursor(1)
             elif keyval == Gdk.KEY_Home:
+                self.__history_search_stop()
                 self.__move_cursor(-10000)
             elif keyval == Gdk.KEY_End:
+                self.__history_search_stop()
                 self.__move_cursor(10000)
             elif keyval == Gdk.KEY_Tab:
+                self.__history_search_stop()
                 cursor = self.__get_cursor()
                 if cursor.starts_line():
                     handled = False
@@ -324,17 +398,32 @@ class _ReadLine(object):
                     else:
                         self.tab_pressed = tab_pressed + 1
                         self.__complete()
+            elif self._searching_backward:
+                if event.string != '' or keyval == Gdk.KEY_BackSpace:
+                  self.__history_search_backward(keyval, event.string)
+            elif self._searching_forward:
+                if event.string != '' or keyval == Gdk.KEY_BackSpace:
+                  self.__history_search_forward(keyval, event.string)
             else:
                 handled = False
         elif state == Gdk.ModifierType.CONTROL_MASK:
             if keyval == Gdk.KEY_u:
+                self.__history_search_stop()
                 start = self.__get_start()
                 end = self.__get_cursor()
                 self.__delete(start, end)
             elif keyval == Gdk.KEY_d:
+                self.__history_search_stop()
                 if self.quit_func:
                     self.quit_func()
+            elif keyval == Gdk.KEY_r:
+                self.__history_search_backward(keyval, '')
+            elif keyval == Gdk.KEY_s:
+                self.__history_search_forward(keyval, '')
+            elif keyval == Gdk.KEY_c and (self._searching_backward or self._searching_forward):
+                self.__history_search_cancel()
             else:
+                self.__history_search_stop()
                 handled = False
         else:
             handled = False
@@ -353,17 +442,101 @@ class _ReadLine(object):
         self.__move_cursor(0)
         self.scroll_to_mark(self.cursor, 0.2, False, 0.0, 0.0)
 
+    def __history_search_stop(self):
+        if self._searching_backward or self._searching_forward:
+          self._searching_backward = False
+          self._searching_forward  = False
+          self.ps = self.ps1
+          self.__replace_line(self.history.items[self.history.ptr])
+          self.__move_cursor(0)
+          self.scroll_to_mark(self.cursor, 0.2, False, 0.0, 0.0)
+
+    def __history_search_cancel(self):
+        if self._searching_backward or self._searching_forward:
+          self._searching_backward = False
+          self._searching_forward  = False
+          self.history.ptr = self._search_initial_ptr
+          self.ps = self.ps1
+          self.__replace_line(self.history.items[self.history.ptr])
+          self.__move_cursor(0)
+          self.scroll_to_mark(self.cursor, 0.2, False, 0.0, 0.0)
+
+    def __history_search_backward(self, keyval, text):
+        search_anyway = False
+
+        if not self._searching_backward:
+          self._searching_backward = True
+          if not self._searching_forward:
+            self._search_initial_ptr = self.history.ptr
+            self._search_text = ''
+          else:
+            search_anyway = True
+          self.ps = self.sb
+        else:
+          if keyval == Gdk.KEY_BackSpace:
+            search_anyway = True
+            self._search_text = self._search_text[:-1]
+          elif text == '':
+            search_anyway = True
+          else:
+            self._search_text += text
+
+        self._searching_forward = False
+
+        if self._search_text != '' and (self._search_text not in self._get_line() or search_anyway):
+          ptr, new_text = self.history.search_backward(self._search_text)
+          if ptr is not None:
+            self.history.ptr = ptr
+
+        self.__replace_line(self.history.items[self.history.ptr])
+
+        self.__move_cursor(0)
+        self.scroll_to_mark(self.cursor, 0.2, False, 0.0, 0.0)
+
+    def __history_search_forward(self, keyval, text):
+        search_anyway = False
+
+        if not self._searching_forward:
+          self._searching_forward = True
+          if not self._searching_backward:
+            self._search_initial_ptr = self.history.ptr
+            self._search_text = ''
+          else:
+            search_anyway = True
+          self.ps = self.sf
+        else:
+          if keyval == Gdk.KEY_BackSpace:
+            search_anyway = True
+            self._search_text   = self._search_text[:-1]
+          elif text == '':
+            search_anyway = True
+          else:
+            self._search_text += text
+
+        self._searching_backward = False
+
+        if self._search_text != '' and (self._search_text not in self._get_line() or search_anyway):
+          ptr, new_text = self.history.search_forward(self._search_text)
+          if ptr is not None:
+            self.history.ptr = ptr
+
+        self.__replace_line(self.history.items[self.history.ptr])
+
+        self.__move_cursor(0)
+        self.scroll_to_mark(self.cursor, 0.2, False, 0.0, 0.0)
+
     def __get_cursor(self):
         '''Returns an iterator at the current cursor position.'''
         return self.buffer.get_iter_at_mark(self.cursor)
 
-    def __get_start(self):
+    def __get_start(self, before_ps=False):
         '''Returns an iterator at the start of the input on the current
         cursor line.'''
 
         iter = self.__get_cursor()
         iter.set_line(iter.get_line())
-        iter.forward_chars(len(self.ps))
+        if not before_ps:
+          iter.forward_chars(len(self.ps))
         return iter
 
     def __get_end(self):
@@ -495,9 +668,12 @@ class _ReadLine(object):
 
     def __replace_line(self, new_text):
         '''Replace the current input with 'new_text' '''
-        start = self.__get_start()
+        start = self.__get_start(True)
         end = self.__get_end()
         self.__delete(start, end)
+        self.__insert(end, self.ps)
+        if self._searching_backward or self._searching_forward:
+          self.__insert(end, '({}) '.format(self._search_text))
         self.__insert(end, new_text)
 
     def _commit(self):
@@ -533,8 +709,9 @@ class _ReadLine(object):
 class _Console(_ReadLine, code.InteractiveInterpreter):
     def __init__(self, locals=None, banner=None,
                  completer=None, use_rlcompleter=True,
-                 start_script=None, quit_func=None):
-        _ReadLine.__init__(self, quit_func)
+                 start_script=None, quit_func=None,
+                 initial_history=None):
+        _ReadLine.__init__(self, quit_func, initial_history)
 
         code.InteractiveInterpreter.__init__(self, locals)
         self.locals["__console__"] = self
@@ -559,6 +736,8 @@ class _Console(_ReadLine, code.InteractiveInterpreter):
 
         self.ps1 = ">>> "
         self.ps2 = "... "
+        self.sb  = "^^^ "
+        self.sf  = "vvv "
         self.__start()
         self.run_on_raw_input = start_script
         self.raw_input(self.ps1)
@@ -603,7 +782,12 @@ class _Console(_ReadLine, code.InteractiveInterpreter):
             ps = self.ps2
         else:
             self.cmd_buffer = ''
-            ps = self.ps1
+            if self._searching_backward:
+                ps = self.sb
+            elif self._searching_forward:
+                ps = self.sb
+            else:
+                ps = self.ps1
 
         sys.stdout, sys.stderr = saved_stdout, saved_stderr
         self.raw_input(ps)
